@@ -27,9 +27,36 @@ __author__ = "Moon Ki Jung"
 import os
 import sys
 import numpy as np
+from scipy.signal import butter, filtfilt
+from scipy.spatial.transform import Rotation as R
 import re
 import btk
 #%%
+def filt_bw_bp(data, fc_low, fc_high, fs, order=2):
+    nyq = 0.5 * fs
+    low = fc_low / nyq
+    high = fc_high / nyq
+    b, a = butter(order, [low, high], analog=False, btype='bandpass', output='ba')
+    axis = -1 if len(data.shape)==1 else 0
+    y = filtfilt(b, a, data, axis, padtype='odd', padlen=3*(max(len(b),len(a))-1))
+    return y
+
+def filt_bw_bs(data, fc_low, fc_high, fs, order=2):
+    nyq = 0.5 * fs
+    low = fc_low / nyq
+    high = fc_high / nyq
+    b, a = butter(order, [low, high], analog=False, btype='bandstop', output='ba')
+    axis = -1 if len(data.shape)==1 else 0
+    y = filtfilt(b, a, data, axis, padtype='odd', padlen=3*(max(len(b),len(a))-1))
+    return y
+
+def filt_bw_lp(data, fc_low, fs, order=2):
+    nyq = 0.5 * fs
+    low = fc_low / nyq
+    b, a = butter(order, low, analog=False, btype='lowpass', output='ba')
+    axis = -1 if len(data.shape)==1 else 0
+    y = filtfilt(b, a, data, axis, padtype='odd', padlen=3*(max(len(b),len(a))-1))
+    return y
 
 def open_c3d(f_path=None):
     if f_path is None: return None
@@ -71,6 +98,42 @@ def get_video_frames(acq):
     frs = np.linspace(first_fr, last_fr, n_frs, dtype=np.int32)
     return frs
 
+def get_analog_frames(acq):
+    first_fr = get_first_frame(acq)
+    last_fr = get_last_frame(acq)  
+    av_ratio = get_analog_video_ratio(acq)
+    start_fr = np.float32(first_fr)
+    end_fr = np.float32(last_fr)+np.float32(av_ratio-1)/np.float32(av_ratio)
+    n_frs = last_fr-first_fr+1
+    analog_steps = n_frs*av_ratio
+    frs = np.linspace(start=start_fr, stop=end_fr, num=analog_steps, dtype=np.float32)
+    return frs
+
+def get_video_times(acq, from_zero=True):
+    first_fr = get_first_frame(acq)
+    last_fr = get_last_frame(acq)
+    vid_fps = get_video_fps(acq)
+    offset_fr = first_fr if from_zero else 0
+    start_t = np.float32(first_fr-offset_fr)/vid_fps
+    end_t = np.float32(last_fr-offset_fr)/vid_fps
+    n_frs = last_fr-first_fr+1
+    t = np.linspace(start=start_t, stop=end_t, num=n_frs, dtype=np.float32)
+    return t
+
+def get_analog_times(acq, from_zero=True):
+    first_fr = get_first_frame(acq)
+    last_fr = get_last_frame(acq)
+    vid_fps = get_video_fps(acq)
+    analog_fps = get_analog_fps(acq)
+    av_ratio = get_analog_video_ratio(acq)
+    offset_fr = first_fr if from_zero else 0
+    start_t = np.float32(first_fr-offset_fr)/vid_fps
+    end_t = np.float32(last_fr-offset_fr)/vid_fps+np.float32(av_ratio-1)/analog_fps
+    vid_steps = last_fr-first_fr+1
+    analog_steps = vid_steps*av_ratio
+    t = np.linspace(start=start_t, stop=end_t, num=analog_steps, dtype=np.float32)
+    return t
+
 def get_point_names(acq, tgt_types=None):
     pt_names = []
     pt_types = None if tgt_types is None else set()
@@ -279,7 +342,7 @@ def get_fp_info(acq):
         dict_fp[i].update({'CAL_MATRIX': np.asarray(cal_matrix.T, dtype=np.float32)})
     return dict_fp
 
-def get_fp_output(acq, threshold=None):
+def get_fp_output(acq, threshold=0.0, filt_fc=None, filt_order=2, cop_nan_to_num=True):
     pfe = btk.btkForcePlatformsExtractor()
     pfe.SetInput(acq)
     pfe.Update()
@@ -288,6 +351,7 @@ def get_fp_output(acq, threshold=None):
         return None
     point_unit = acq.GetPointUnit()
     point_scale = 1.0 if point_unit=='m' else 0.001
+    analog_fps = acq.GetAnalogFrequency()
     rgx_fp = re.compile(r'\S*(\d*[FMP]\d*[XxYyZz]\d*)')
     fp_output = {}
     fp_idx = 0
@@ -337,6 +401,17 @@ def get_fp_output(acq, threshold=None):
         ch_scale = {}
         # for ch in btk.Iterate(fp.GetChannels()):
         fp_cnt_chs = fp.GetChannelNumber()
+        if filt_fc is None:
+            filt_fcs = [None]*fp_cnt_chs
+        elif type(filt_fc) in [int, float]:
+            filt_fcs = [float(filt_fc)]*fp_cnt_chs
+        elif type(filt_fc) in [list, tuple]:
+            filt_fcs = list(filt_fc)
+        elif type(filt_fc)==np.ndarray:
+            if len(filt_fc)==1:
+                filt_fcs = [filt_fc.item()]*fp_cnt_chs
+            else:
+                filt_fcs = filt_fc.tolist()
         for ch_idx in range(fp_cnt_chs):
             ch_name = fp.GetChannel(ch_idx).GetLabel()
             ch = acq.GetAnalog(ch_name)
@@ -374,7 +449,11 @@ def get_fp_output(acq, threshold=None):
                     ch_scale[label] = 0.001
                     if ch.GetUnit()=='Nm': ch_scale[label] = 1.0
             # assign channel values
-            ch_data[label] = np.squeeze(ch.GetData().GetValues())
+            lp_fc = filt_fcs[ch_idx]
+            if lp_fc is None:
+                ch_data[label] = np.squeeze(ch.GetData().GetValues())
+            else:
+                ch_data[label] = filt_bw_lp(np.squeeze(ch.GetData().GetValues()), lp_fc, analog_fps, order=filt_order)
         if fp_type == 1:
             cop_l_x_in = ch_data['PX']*ch_scale['PX']
             cop_l_y_in = ch_data['PY']*ch_scale['PY']
@@ -413,11 +492,8 @@ def get_fp_output(acq, threshold=None):
             mz = fp_len_b*(-fx12+fx34)+fp_len_a*(fy14-fy23)
             f_raw = np.stack([fx, fy, fz], axis=1)
             m_raw = np.stack([mx, my, mz], axis=1)
-        if threshold is None:
-            fm_skip_mask = np.full((f_raw.shape[0],), False, dtype=bool)
-        else:
-            fm_skip_mask = np.abs(f_raw[:,2])<=threshold
         zero_vals = np.zeros((f_raw.shape[0]), dtype=np.float32)
+        fm_skip_mask = np.abs(f_raw[:,2])<=threshold
         f_sensor_local = f_raw.copy()
         m_sensor_local = m_raw.copy()
         # filter local values by threshold
@@ -430,11 +506,14 @@ def get_fp_output(acq, threshold=None):
         m_y = m_sensor_local[:,1]
         m_z = m_sensor_local[:,2]
         with np.errstate(invalid='ignore'):
-            # cop_l_x = np.nan_to_num(np.where(fm_skip_mask, 0, (-m_y+(-o_z)*f_x)/f_z+o_x))
-            # cop_l_y = np.nan_to_num(np.where(fm_skip_mask, 0, (m_x+(-o_z)*f_y)/f_z+o_y)) 
-            cop_l_x = np.clip(np.nan_to_num(np.where(fm_skip_mask, 0, (-m_y+(-o_z)*f_x)/f_z+o_x)), -fp_len_x*0.5, fp_len_x*0.5)
-            cop_l_y = np.clip(np.nan_to_num(np.where(fm_skip_mask, 0, (m_x+(-o_z)*f_y)/f_z+o_y)), -fp_len_y*0.5, fp_len_y*0.5)
-            cop_l_z = zero_vals
+            f_z_adj = np.where(fm_skip_mask, np.inf, f_z)          
+            cop_l_x = np.where(fm_skip_mask, np.nan, np.clip((-m_y+(-o_z)*f_x)/f_z_adj+o_x, -fp_len_x*0.5, fp_len_x*0.5))
+            cop_l_y = np.where(fm_skip_mask, np.nan, np.clip((m_x+(-o_z)*f_y)/f_z_adj+o_y, -fp_len_y*0.5, fp_len_y*0.5))
+            cop_l_z = np.where(fm_skip_mask, np.nan, zero_vals)
+            if cop_nan_to_num:
+                cop_l_x = np.nan_to_num(cop_l_x)
+                cop_l_y = np.nan_to_num(cop_l_y)
+                cop_l_z = np.nan_to_num(cop_l_z)
         t_z = m_z-(cop_l_x-o_x)*f_y+(cop_l_y-o_y)*f_x
         # values for the force plate local output
         m_cop_local = np.stack([zero_vals, zero_vals, t_z], axis=1)
@@ -516,4 +595,95 @@ def get_metadata(acq, parent_name, name):
 def change_point_name(acq, old_name, new_name):
     pt = acq.FindPoint(old_name).value()
     pt.SetLabel(new_name)
-    
\ No newline at end of file
+    
+def export_trc(acq, f_path, rot_mat=np.eye(3), filt_fc=None, filt_order=2, tgt_mkr_names=None, start_fr=None, end_fr=None, fmt='%.6f'):
+    dict_pts = get_dict_points(acq, tgt_types=['Marker'])
+    orig_start_fr = 1
+    vid_fps = int(get_video_fps(acq))
+    if start_fr is None:
+        start_fr = get_first_frame(acq)
+    if end_fr is None:
+        end_fr = get_last_frame(acq)
+    n_vid_frs = end_fr-start_fr+1
+    vid_frs = np.linspace(start_fr, end_fr, n_vid_frs, dtype=np.int32)
+    vid_times = (vid_frs-orig_start_fr)/vid_fps
+    orig_vid_frs = get_video_frames(acq)
+    frs_sel_mask = np.logical_and(orig_vid_frs>=vid_frs[0], orig_vid_frs<=vid_frs[-1])
+    total_mkr_names = list(dict_pts['LABELS'])
+    if tgt_mkr_names is None:
+        mkr_names = total_mkr_names
+    else:
+        mkr_names = [x for x in list(tgt_mkr_names) if x in total_mkr_names]
+    mkr_unit = 'mm'
+    hdr_row0 = f'PathFileType\t4\t(X/Y/Z)\t{os.path.basename(f_path)}'
+    hdr_row1 = 'DataRate\tCameraRate\tNumFrames\tNumMarkers\tUnits\tOrigDataRate\tOrigDataStartFrame\tOrigNumFrames'
+    hdr_row2 = f'{vid_fps}\t{vid_fps}\t{n_vid_frs}\t{len(mkr_names)}\t{mkr_unit}\t{vid_fps}\t{orig_start_fr}\t{orig_vid_frs.shape[0]}'
+    hdr_row3 = 'Frame#\tTime\t'+'\t\t\t'.join(mkr_names)+'\t\t'
+    hdr_row4 = '\t\t'+'\t'.join([c+str(n) for n in list(range(1, len(mkr_names)+1)) for c in ['X', 'Y', 'Z']])
+    hdr_row5 = '\t'
+    hdr_str = '\n'.join([hdr_row0, hdr_row1, hdr_row2, hdr_row3, hdr_row4, hdr_row5])
+    output_data = np.zeros((n_vid_frs, 2+3*len(mkr_names)), dtype=float)
+    output_data[:,0] = vid_frs
+    output_data[:,1] = vid_times
+    for mkr_idx, mkr_name in enumerate(mkr_names):
+        mkr_pos_raw = np.dot(rot_mat, dict_pts['DATA']['POS'][mkr_name].T).T
+        if filt_fc is None:
+            mkr_pos = mkr_pos_raw
+        else:
+            mkr_pos = filt_bw_lp(mkr_pos_raw, filt_fc, vid_fps, order=filt_order)
+        output_data[:,2+3*mkr_idx:2+3*mkr_idx+3] = mkr_pos[frs_sel_mask,:]
+    fmt_str = f'%d\t{fmt}\t'+'\t'.join([fmt]*(3*len(mkr_names)))
+    np.savetxt(f_path, output_data, fmt=fmt_str, delimiter='\t', comments='', header=hdr_str)
+    
+def export_mot(acq, f_path, rot_mat=np.eye(3), filt_fc=None, filt_order=2, threshold=0.0, start_fr=None, end_fr=None, fmt='%.6f'):
+    orig_start_fr = 1
+    if start_fr is None:
+        start_fr = get_first_frame(acq)
+    if end_fr is None:
+        end_fr = get_last_frame(acq) 
+    vid_fps = int(get_video_fps(acq))
+    anal_fps = get_analog_fps(acq)
+    av_ratio = get_analog_video_ratio(acq)
+    n_vid_frs = end_fr-start_fr+1
+    vid_frs = np.linspace(start_fr, end_fr, n_vid_frs, dtype=np.int32)
+    vid_times = (vid_frs-orig_start_fr)/vid_fps
+    first_fr = get_first_frame(acq)
+    last_fr = get_last_frame(acq)
+    anal_start_t = np.float32(first_fr-orig_start_fr)/vid_fps
+    anal_end_t = np.float32(last_fr-orig_start_fr)/vid_fps+np.float32(av_ratio-1)/anal_fps
+    anal_steps = (last_fr-first_fr+1)*av_ratio
+    anal_times = np.linspace(start=anal_start_t, stop=anal_end_t, num=anal_steps, dtype=np.float32)
+    anal_sel_mask = np.logical_and(anal_times>vid_times[0]-(1.0/anal_fps), anal_times<vid_times[-1]+(1.0/anal_fps))
+    anal_sel_times = anal_times[anal_sel_mask]
+    # rot_ret = R.align_vectors(a=csys_src, b=csys_tgt)
+    # rot_obj = rot_ret[0]
+    fp_output = get_fp_output(acq, threshold=threshold, filt_fc=filt_fc, filt_order=filt_order)
+    cnt_fp = len(fp_output)
+    output_col_names = []
+    for i in range(cnt_fp):
+        output_col_names.append(f'ground_force{i+1}_vx')
+        output_col_names.append(f'ground_force{i+1}_vy')
+        output_col_names.append(f'ground_force{i+1}_vz')        
+        output_col_names.append(f'ground_force{i+1}_px')
+        output_col_names.append(f'ground_force{i+1}_py')
+        output_col_names.append(f'ground_force{i+1}_pz')
+        output_col_names.append(f'ground_torque{i+1}_x')
+        output_col_names.append(f'ground_torque{i+1}_y')
+        output_col_names.append(f'ground_torque{i+1}_z')
+    hdr_row0 = f'{os.path.basename(f_path)}'
+    hdr_row1 = f'datacolumns\t{1+9*cnt_fp}'
+    hdr_row2 = f'datarows\t{anal_sel_times.shape[0]}'
+    hdr_row3 = f'range\t{anal_sel_times[0]:{fmt.replace("%","")}}\t{anal_sel_times[-1]:{fmt.replace("%","")}}'
+    hdr_row4 = 'endheader'
+    hdr_row5 = 'time\t'+'\t'.join(output_col_names)
+    hdr_str = '\n'.join([hdr_row0, hdr_row1, hdr_row2, hdr_row3, hdr_row4, hdr_row5])
+    output_data = np.zeros((anal_sel_times.shape[0], len(output_col_names)+1), dtype=float)
+    output_data[:,0] = anal_sel_times
+    for i in range(cnt_fp):
+        cop_lab = fp_output[i]['COP_LAB'][anal_sel_mask,:]
+        f_cop_lab = fp_output[i]['F_COP_LAB'][anal_sel_mask,:]
+        m_cop_lab = fp_output[i]['M_COP_LAB'][anal_sel_mask,:]
+        output_data[:,9*i+1:9*i+4] = np.dot(rot_mat, f_cop_lab.T).T
+        output_data[:,9*i+4:9*i+7] = np.dot(rot_mat, cop_lab.T).T
+        output_data[:,9*i+7:9*i+10] = np.dot(rot_mat, m_cop_lab.T).T
+    np.savetxt(f_path, output_data, fmt=fmt, delimiter='\t', comments='', header=hdr_str)
\ No newline at end of file
